from clovers_agent import Event, CloversAgent
from .toolkit import toolkit
from .config import __config__

BRAVE_API_KEY = __config__.BRAVE_API_KEY


@toolkit.tool(
    "web_search",
    "联网搜索",
    {"query": {"type": "string", "description": "搜索关键词"}},
    ["从URL获取资源"],
)
async def _(agent: CloversAgent, event: Event, query: list[str]):
    headers = {"Accept": "application/json", "Accept-Encoding": "gzip", "X-Subscription-Token": BRAVE_API_KEY}
    url = "https://api.search.brave.com/res/v1/web/search"
    params = {"q": query, "count": 8}
    resp = await agent.async_client.get(url, headers=headers, params=params, timeout=20.0)
    if resp.status_code != 200:
        return f"搜索失败，状态码：{resp.status_code}"
    try:
        results = resp.json()["web"]["results"]
    except KeyError:
        return "服务器错误，请稍后再试。"
    if not results:
        return f"未找到关于 '{query}' 的相关搜索结果。"
    md_output = [f"### 关于 '{query}' 的搜索结果：\n"]
    for idx, item in enumerate(results, 1):
        title = item.get("title", "无标题")
        link = item.get("url", "#")
        snippet = item.get("description", "无摘要")
        md_output.append(f"{idx}. **[{title}]({link})**\n   摘要: {snippet}\n")
    return "\n".join(md_output)


@toolkit.tool(
    "web_extractor",
    "读取指定 URL 的网页纯文本内容。当需要从特定网页获取文本信息时使用。",
    {"webpage_url": {"type": "string", "description": "网页的完整 URL 地址"}},
    ["从URL获取资源"],
)
async def _(agent: CloversAgent, event: Event, webpage_url: str):
    if not webpage_url.startswith("http"):
        webpage_url = f"https://{webpage_url}"
    try:
        resp = await agent.async_client.get(webpage_url)
        if resp.status_code != 200:
            return f"获取网页失败，状态码：{resp.status_code}"
        return resp.text
    except Exception:
        return "获取网页失败"


@toolkit.tool(
    "view_image_url",
    "查看网络图片。当你需要查看用户提供的图片链接时，请调用此工具",
    {"image_url": {"type": "string", "description": "图片的完整 URL 地址"}},
    ["从URL获取资源"],
)
async def _(agent: CloversAgent, event: Event, image_url: str):
    if not image_url.startswith("http"):
        image_url = f"https://{image_url}"
    assert agent.current_input
    if isinstance(agent.current_input["content"], str):
        agent.current_input["content"] = [{"type": "text", "text": agent.current_input["content"]}]
    agent.current_input["content"].append({"type": "image_url", "image_url": {"url": image_url}})
    return "图片已放入用户上下文"



from clovers_agent import Event, CloversAgent
from clovers.logger import logger
from .docker import WORKSPACE, Shell
from ..toolkit import toolkit
from ..config import __config__

README = WORKSPACE / "README.md"
shell_dict: dict[str, Shell] = {}


def get_session_id(agent: CloversAgent, event: Event) -> str: ...


if __config__.session_workspace:
    get_session_id = lambda agent, event: agent.session_id(event)
else:
    get_session_id = lambda agent, event: "public"


@toolkit.on_skill("工作区工具")
async def _(agent: CloversAgent, event: Event):
    if not WORKSPACE.exists():
        WORKSPACE.mkdir(parents=True, exist_ok=True)
    if not README.exists():
        README.write_text("Clovers Agent Workspace")
    if __config__.use_shell:
        session_id = get_session_id(agent, event)
        if session_id in shell_dict:
            shell_dict[session_id].workdir = "/workspace"
        else:
            try:
                shell_dict[session_id] = Shell(session_id)
            except Exception as e:
                logger.error(e)
                return f"workspace 已初始化, shell 初始化失败:{e}\n当前工作目录: /workspace"
        return f"workspace 已初始化，当前系统：Debian\n当前工作目录: /workspace"
    else:
        return f"workspace 已初始化\n当前工作目录: /workspace"


if __config__.use_shell:

    @toolkit.tool(
        "shell",
        "在工作区环境下执行命令",
        {"command": {"type": "string", "description": "需要执行的命令，如需要执行多条命令，请使用 `&&` 或 `;`隔开"}},
        ["工作区工具"],
    )
    async def _(agent: CloversAgent, event: Event, command: str):
        session_id = get_session_id(agent, event)
        if session_id not in shell_dict:
            return f"Error: shell 初始化失败，请返回故障原因。在故障排除前不要重复调用此方法。"
        shell = shell_dict[session_id]
        logger.info(f"[CloversAgentShell][{session_id}]> {command if (idx := command.find("\n")) == -1 else f"{command[:idx]}..."}")
        output = await shell.execute(command)
        return f"{output}\n当前工作目录: {shell.workdir}"


@toolkit.tool(
    "read_files",
    "读取并查看指定文件的内容。支持同时传入多个路径以一次性查看多个文件上下文。"
    "在需要分析代码、检查配置文件时，尤其是需要查看多个文件时，应优先使用此工具以提高效率。",
    {"filepaths": {"type": "array", "description": "包含一个或多个文件路径的数组", "items": {"type": "string"}}},
    ["工作区工具"],
)
async def _(agent: CloversAgent, event: Event, filepaths: list[str]):
    session_id = get_session_id(agent, event)
    workspace = WORKSPACE / session_id
    md = []
    for file_path in filepaths:
        if file_path.startswith("/workspace"):
            file = workspace / f"./{file_path[10:]}"
        else:
            file = workspace / file_path
        if not file.is_relative_to(workspace):
            md.append(f"{file_path} 非工作区文件")
        if not file.exists():
            md.append(f"{file_path} 文件不存在")
            continue
        try:
            content = file.read_text()
        except Exception as e:
            logger.error(f"{file_path} 文件读取失败:{e}")
            md.append(f"{file_path} 文件读取失败")
            continue
        md.append(f"```{file_path}\n{content}\n```")
    return "\n\n".join(md)


@toolkit.tool(
    "write_file",
    "写入文件",
    {
        "file_path": {"type": "string", "description": "需要写入的文件路径"},
        "file_content": {"type": "string", "description": "需要写入到文件的内容"},
    },
    ["工作区工具"],
)
async def _(agent: CloversAgent, event: Event, file_path: str, file_content: str):
    session_id = get_session_id(agent, event)
    workspace = WORKSPACE / session_id
    if file_path.startswith("/workspace"):
        file = workspace / f"./{file_path[10:]}"
    else:
        file = workspace / file_path
    file.parent.mkdir(parents=True, exist_ok=True)
    try:
        file.write_text(file_content, encoding="utf-8")
        return f"文件写入成功。"
    except Exception as e:
        logger.error(e)
        return f"文件写入失败：{e}"



from pydantic import BaseModel
from clovers.config import Config as CloversConfig


class Config(BaseModel):
    BRAVE_API_KEY: str
    use_shell: bool = True
    """是否使用shell"""
    session_workspace: bool = True
    """为每个会话创建一个工作空间"""

    @classmethod
    def sync_config(cls):
        """获取 `CloversConfig.environ()[__package__]` 配置并将默认配置同步到全局配置中。"""
        __config_dict__: dict = CloversConfig.environ().setdefault(__package__, {})
        __config_dict__.update((__config__ := cls.model_validate(__config_dict__)).model_dump())
        return __config__


__config__ = Config.sync_config()



import asyncio
import docker
import shlex
from docker.models.containers import Container
from docker.types import DeviceRequest
from docker.errors import NotFound
from pathlib import Path


WORKSPACE = Path("workspace")
client: docker.DockerClient | None = None


class Shell:

    def __init__(self, session_id: str):
        self.lock = asyncio.Lock()
        self.session_id = session_id
        self.workdir = "/workspace"
        self.container: Container | None = None
        global client
        if client is None:
            client = docker.from_env()
        self.client = client

    async def execute(self, command: str):
        async with self.lock:
            if self.container is None:
                self.workdir = "/workspace"
                workspace = WORKSPACE / self.session_id
                if not workspace.exists():
                    workspace.mkdir(parents=True, exist_ok=True)
                container_name = f"CloversAgentSandbox-{self.session_id}"
                try:
                    self.container = await asyncio.to_thread(self.client.containers.get, container_name)
                except NotFound:
                    self.container = await asyncio.to_thread(
                        self.client.containers.run,
                        "nikolaik/python-nodejs:python3.12-nodejs20",
                        name=container_name,
                        detach=True,
                        tty=True,
                        command="sleep infinity",
                        volumes={workspace.resolve().as_posix(): {"bind": "/workspace", "mode": "rw"}},
                        device_requests=[DeviceRequest(count=-1, capabilities=[["gpu"]])],
                    )
            else:
                self.container
                self.container.reload()
            if self.container.status != "running":
                await asyncio.to_thread(self.container.start)
            assert self.container is not None
            wrapped_command = f"bash -c {shlex.quote(f"{command}\necho '___CWD_MARKER___'\npwd")}"
            # result = await asyncio.to_thread(self.container.exec_run, wrapped_command, workdir=self.workdir)
            # stdout: str = result.output.decode("utf-8")
            stdout = await asyncio.to_thread(self.execute_thread, wrapped_command)
            output, workdir = stdout.rsplit("___CWD_MARKER___", 1)
            self.workdir = workdir.strip()
            return output

    async def cleanup(self):
        async with self.lock:
            if self.container is None:
                return
            await asyncio.to_thread(self.container.remove, force=True)
            self.workdir = "/workspace"
            self.container = None

    def execute_thread(self, command: str):
        """运行命令,不输出被回车覆盖的行"""
        exec_id = self.client.api.exec_create(self.container.id, command, workdir=self.workdir)  # type:ignore
        output_gen = self.client.api.exec_start(exec_id["Id"], stream=True)
        outputs: list[str] = []
        buffer: bytearray = bytearray()
        for chunk in output_gen:
            for byte in chunk:
                buffer.append(byte)
                if byte == 10:
                    line = buffer.decode("utf-8")
                    # print(line)
                    outputs.append(line)
                    buffer.clear()
                elif byte == 13:
                    # sys.stdout.buffer.write(buffer)
                    # sys.stdout.flush()
                    buffer.clear()
        if buffer:
            outputs.append(buffer.decode("utf-8"))
        return "\n".join(outputs)
